/*
 ## Cypress USB 3.0 Platform source file (cyfxslfifosync.c)
 ## ===========================
 ##
 ##  Copyright Cypress Semiconductor Corporation, 2010-2011,
 ##  All Rights Reserved
 ##  UNPUBLISHED, LICENSED SOFTWARE.
 ##
 ##  CONFIDENTIAL AND PROPRIETARY INFORMATION
 ##  WHICH IS THE PROPERTY OF CYPRESS.
 ##
 ##  Use of this file is governed
 ##  by the license agreement included in the file
 ##
 ##     <install>/license/license.txt
 ##
 ##  where <install> is the Cypress software
 ##  installation root directory path.
 ##
 ## ===========================
*/

/* This file illustrates the Slave FIFO Synchronous mode example */

/*
   This example comprises of two USB bulk endpoints. A bulk OUT endpoint acts as the
   producer of data from the host. A bulk IN endpoint acts as the consumer of data to
   the host. Appropriate vendor class USB enumeration descriptors with these two bulk
   endpoints are implemented.

   The GPIF configuration data for the Synchronous Slave FIFO operation is loaded onto
   the appropriate GPIF registers. The p-port data transfers are done via the producer
   p-port socket and the consumer p-port socket.

   This example implements two DMA Channels in MANUAL mode one for P to U data transfer
   and one for U to P data transfer.

   The U to P DMA channel connects the USB producer (OUT) endpoint to the consumer p-port
   socket. And the P to U DMA channel connects the producer p-port socket to the USB 
   consumer (IN) endpoint.

   Upon every reception of data in the DMA buffer from the host or from the p-port, the
   CPU is signalled using DMA callbacks. There are two DMA callback functions implemented
   each for U to P and P to U data paths. The CPU then commits the DMA buffer received so
   that the data is transferred to the consumer.

   The DMA buffer size for each channel is defined based on the USB speed. 64 for full
   speed, 512 for high speed and 1024 for super speed. CY_FX_SLFIFO_DMA_BUF_COUNT in the
   header file defines the number of DMA buffers per channel.

   The constant CY_FX_SLFIFO_GPIF_16_32BIT_CONF_SELECT in the header file is used to
   select 16bit or 32bit GPIF data bus configuration.
 */

#include "cyu3system.h"
#include "cyu3os.h"
#include "cyu3dma.h"
#include "cyu3error.h"
#include "cyu3usb.h"
#include "cyu3uart.h"
#include "cyfxslfifosync.h"
#include "cyu3gpif.h"
#include "cyu3pib.h"
#include "pib_regs.h"

/* This file should be included only once as it contains
 * structure definitions. Including it in multiple places
 * can result in linker error. */
#ifdef EXPLORE_GPIF_NOISE
#include "FX3gpif_selfRDWR.h"
#else
#include "cyfxgpif_syncsf.h"
#endif


#ifdef EXPLORE_GPIF_NOISE
static volatile uint32_t glPhyErrorCount = 0;
static volatile uint32_t glLnkErrorCount = 0;
static volatile uint32_t glUSBDisconnectCount = 0;
static volatile uint32_t glErrorUpdateCount=0;
static volatile uint32_t glLinkErrLimitExceeded=0;
static volatile uint32_t glLinkRecovery=0;
#endif

#if 10
static uint32_t read_CPSR(void)
{
	register uint32_t cpsr;
    __asm__ volatile(
        "MRS %0, CPSR\n\t"
       : "=r"(cpsr)
    );
    return cpsr;
}
static void write_CPSR(register uint32_t cpsr)
{
    __asm__ volatile(
        "MSR CPSR, %0;"
		:
        : "r"(cpsr)
    );
}

/* Set I and F bits in Program Status Register, i.e. disable IRQ and FIQ interrupts */
static uint32_t disable_interrupts( void )
{
	uint32_t cpsr=read_CPSR();
	write_CPSR( cpsr | 0xC0 );
	return cpsr;
}

/* Restore I and F bits in Program Status Register  */
static void restore_interrupts( register uint32_t cpsr )
{
	write_CPSR( (cpsr & 0xC0) | (read_CPSR() & ~0xC0) );
}
#endif


CyU3PThread slFifoAppThread;	        /* Slave FIFO application thread structure */
CyU3PDmaChannel glChHandleSlFifoUtoP;   /* DMA Channel handle for U2P transfer. */
CyU3PDmaChannel glChHandleSlFifoPtoU;   /* DMA Channel handle for P2U transfer. */

uint32_t glDMARxCount = 0;               /* Counter to track the number of buffers received from USB. */
uint32_t glDMATxCount = 0;               /* Counter to track the number of buffers sent to USB. */
CyBool_t glIsApplnActive = CyFalse;      /* Whether the loopback application is active or not. */

/* Application Error Handler */
void
CyFxAppErrorHandler (
        CyU3PReturnStatus_t apiRetStatus    /* API return status */
        )
{
    /* Application failed with the error code apiRetStatus */

    /* Add custom debug or recovery actions here */

    /* Loop Indefinitely */
    for (;;)
    {
        /* Thread sleep : 100 ms */
        CyU3PThreadSleep (100);
    }
}

/* This function initializes the debug module. The debug prints
 * are routed to the UART and can be seen using a UART console
 * running at 115200 baud rate. */
void
CyFxSlFifoApplnDebugInit (void)
{
    CyU3PUartConfig_t uartConfig;
    CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;

    /* Initialize the UART for printing debug messages */
    apiRetStatus = CyU3PUartInit();
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        /* Error handling */
    	// Error - OK, we just can't debug.
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Set UART configuration */
    CyU3PMemSet ((uint8_t *)&uartConfig, 0, sizeof (uartConfig));
    uartConfig.baudRate = CY_U3P_UART_BAUDRATE_115200;
    uartConfig.stopBit = CY_U3P_UART_ONE_STOP_BIT;
    uartConfig.parity = CY_U3P_UART_NO_PARITY;
    uartConfig.txEnable = CyTrue;
    uartConfig.rxEnable = CyFalse;
    uartConfig.flowCtrl = CyFalse;
    uartConfig.isDma = CyTrue;

    apiRetStatus = CyU3PUartSetConfig (&uartConfig, NULL);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
    	// Error - OK, we just can't debug.
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Set the UART transfer to a really large value. */
    apiRetStatus = CyU3PUartTxSetBlockXfer (0xFFFFFFFF);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
       // Error - OK, we just can't debug.
       // CyFxAppErrorHandler(apiRetStatus);
    }

    /* Initialize the debug module. */
    apiRetStatus = CyU3PDebugInit (CY_U3P_LPP_SOCKET_UART_CONS, 8);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
    	// Error - OK, we just can't debug.
        //CyFxAppErrorHandler(apiRetStatus);
    }
    CyU3PDebugPrint (4, "Debugging initialized\r\n");
}

/* DMA callback function to handle the produce events for U to P transfers. */
void
CyFxSlFifoUtoPDmaCallback (
        CyU3PDmaChannel   *chHandle,
        CyU3PDmaCbType_t  type,
        CyU3PDmaCBInput_t *input
        )
{
    //CyU3PReturnStatus_t status = CY_U3P_SUCCESS;

    if (type == CY_U3P_DMA_CB_PROD_EVENT)
    {
#ifdef EXPLORE_GPIF_NOISE
#ifdef DISCARD_OUT_DATA
    	// discard data so that it's not output to GPIF pins
    	CyU3PDmaChannelDiscardBuffer(chHandle);
#endif

#else  //EXPLORE_GPIF_NOISE

        /* This is a produce event notification to the CPU. This notification is 
         * received upon reception of every buffer. The buffer will not be sent
         * out unless it is explicitly committed. The call shall fail if there
         * is a bus reset / usb disconnect or if there is any application error. */
    	{
    		CyU3PReturnStatus_t status = CyU3PDmaChannelCommitBuffer (chHandle, input->buffer_p.count, 0);
			if (status != CY_U3P_SUCCESS)
			{
				CyU3PDebugPrint (4, "CyU3PDmaChannelCommitBuffer failed, Error code = %d\n", status);
			}
    	}
#endif //EXPLORE_GPIF_NOISE

        /* Increment the counter. */
        glDMARxCount++;
    }
}

/* DMA callback function to handle the produce events for P to U transfers. */
void
CyFxSlFifoPtoUDmaCallback (
        CyU3PDmaChannel   *chHandle,
        CyU3PDmaCbType_t  type,
        CyU3PDmaCBInput_t *input
        )
{
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;

    if (type == CY_U3P_DMA_CB_PROD_EVENT)
    {
        /* This is a produce event notification to the CPU. This notification is 
         * received upon reception of every buffer. The buffer will not be sent
         * out unless it is explicitly committed. The call shall fail if there
         * is a bus reset / usb disconnect or if there is any application error. */
        status = CyU3PDmaChannelCommitBuffer (chHandle, input->buffer_p.count, 0);
        if (status != CY_U3P_SUCCESS)
        {
            CyU3PDebugPrint (4, "CyU3PDmaChannelCommitBuffer failed, Error code = %d\n", status);
        }

        /* Increment the counter. */
        glDMATxCount++;
    }
}

/* This function starts the slave FIFO loop application. This is called
 * when a SET_CONF event is received from the USB host. The endpoints
 * are configured and the DMA pipe is setup in this function. */
void
CyFxSlFifoApplnStart (
        void)
{
    uint16_t size = 0;
    CyU3PEpConfig_t epCfg;
    CyU3PDmaChannelConfig_t dmaCfg;
    CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;
    CyU3PUSBSpeed_t usbSpeed = CyU3PUsbGetSpeed();

    /* First identify the usb speed. Once that is identified,
     * create a DMA channel and start the transfer on this. */

    /* Based on the Bus Speed configure the endpoint packet size */
    switch (usbSpeed)
    {
        case CY_U3P_FULL_SPEED:
            size = 64;
            break;

        case CY_U3P_HIGH_SPEED:
            size = 512;
            break;

        case  CY_U3P_SUPER_SPEED:
            size = 1024;
            break;

        default:
            CyU3PDebugPrint (4, "Error! Invalid USB speed.\n");
            //CyFxAppErrorHandler (CY_U3P_ERROR_FAILURE);
            return;
    }

    CyU3PMemSet ((uint8_t *)&epCfg, 0, sizeof (epCfg));
    epCfg.enable = CyTrue;
    epCfg.epType = CY_U3P_USB_EP_BULK;
    epCfg.burstLen = (CY_U3P_SUPER_SPEED==usbSpeed)?CY_U3P_MAX(1,CY_U3P_MIN(16,EP_PRODUCER_BURST_LENGTH)):0;
    epCfg.streams = 0;
    epCfg.pcktSize = size;

    /* Producer endpoint configuration */
    apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_PRODUCER, &epCfg);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "Start Producer CyU3PSetEpConfig failed, Error code = %d\n", apiRetStatus);
      //  CyFxAppErrorHandler (apiRetStatus);
    }

    /* Consumer endpoint configuration */
    epCfg.burstLen = (CY_U3P_SUPER_SPEED==usbSpeed)?CY_U3P_MAX(1,CY_U3P_MIN(16,EP_PRODUCER_BURST_LENGTH)):0;
    apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_CONSUMER, &epCfg);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "Start Consumer CyU3PSetEpConfig failed, Error code = %d\n", apiRetStatus);
      //  CyFxAppErrorHandler (apiRetStatus);
    }

#ifdef EXPLORE_GPIF_NOISE
    /* For maximal performance, let's create DMA AUTO channels for U2P and P2U transfers.
     * For max burst length 16, at least 16KB buffer should be allocated.
     * We should have enough memory, so let's allocate 32K buffers.
     * NB! FX3 API uses internally 16-bit values for "size", therefore
     * 64K and bigger buffer sizes can't be specified */

    /* Create a DMA channel for U2P transfer. */
    dmaCfg.size  = 1024*32;
    dmaCfg.count = CY_FX_SLFIFO_DMA_BUF_COUNT;
    dmaCfg.prodSckId = CY_FX_PRODUCER_USB_SOCKET;
    dmaCfg.consSckId = CY_FX_CONSUMER_PPORT_SOCKET;
    dmaCfg.dmaMode = CY_U3P_DMA_MODE_BYTE;
    /* For max performance, no callbacks. */
    dmaCfg.notification = 0;
    dmaCfg.cb = 0;
    dmaCfg.prodHeader = 0;
    dmaCfg.prodFooter = 0;
    dmaCfg.consHeader = 0;
    dmaCfg.prodAvailCount = 0;

#ifdef DISCARD_OUT_DATA
    // if we need to discard data, do it in callback
    dmaCfg.notification = CY_U3P_DMA_CB_PROD_EVENT;
    dmaCfg.cb = CyFxSlFifoUtoPDmaCallback;
#endif

    /* Create a DMA AUTO (or DMA MANUAL if we discard data) channel for U2P transfer. */
    apiRetStatus = CyU3PDmaChannelCreate (&glChHandleSlFifoUtoP,
    		dmaCfg.cb? CY_U3P_DMA_TYPE_MANUAL:CY_U3P_DMA_TYPE_AUTO, &dmaCfg);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PDmaChannelCreate failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Create a DMA AUTO channel for P2U transfer. */
    dmaCfg.prodSckId = CY_FX_PRODUCER_PPORT_SOCKET;
    dmaCfg.consSckId = CY_FX_CONSUMER_USB_SOCKET;
    dmaCfg.notification = 0;
    dmaCfg.cb = 0;
    apiRetStatus = CyU3PDmaChannelCreate (&glChHandleSlFifoPtoU,
    		CY_U3P_DMA_TYPE_AUTO, &dmaCfg);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PDmaChannelCreate failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

#else

    /* Create a DMA MANUAL channel for U2P transfer.
     * DMA size is set based on the USB speed. */
    dmaCfg.size  = size;
    dmaCfg.count = CY_FX_SLFIFO_DMA_BUF_COUNT;
    dmaCfg.prodSckId = CY_FX_PRODUCER_USB_SOCKET;
    dmaCfg.consSckId = CY_FX_CONSUMER_PPORT_SOCKET;
    dmaCfg.dmaMode = CY_U3P_DMA_MODE_BYTE;
    /* Enabling the callback for produce event. */
    dmaCfg.notification = CY_U3P_DMA_CB_PROD_EVENT;
    dmaCfg.cb = CyFxSlFifoUtoPDmaCallback;
    dmaCfg.prodHeader = 0;
    dmaCfg.prodFooter = 0;
    dmaCfg.consHeader = 0;
    dmaCfg.prodAvailCount = 0;

    apiRetStatus = CyU3PDmaChannelCreate (&glChHandleSlFifoUtoP,
            CY_U3P_DMA_TYPE_MANUAL, &dmaCfg);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PDmaChannelCreate failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Create a DMA MANUAL channel for P2U transfer. */
    dmaCfg.prodSckId = CY_FX_PRODUCER_PPORT_SOCKET;
    dmaCfg.consSckId = CY_FX_CONSUMER_USB_SOCKET;
    dmaCfg.cb = CyFxSlFifoPtoUDmaCallback;
    apiRetStatus = CyU3PDmaChannelCreate (&glChHandleSlFifoPtoU,
            CY_U3P_DMA_TYPE_MANUAL, &dmaCfg);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PDmaChannelCreate failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }
#endif

    /* Flush the Endpoint memory */
    CyU3PUsbFlushEp(CY_FX_EP_PRODUCER);
    CyU3PUsbFlushEp(CY_FX_EP_CONSUMER);

    /* Set DMA channel transfer size. */
    apiRetStatus = CyU3PDmaChannelSetXfer (&glChHandleSlFifoUtoP, CY_FX_SLFIFO_DMA_TX_SIZE);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PDmaChannelSetXfer Failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }
    apiRetStatus = CyU3PDmaChannelSetXfer (&glChHandleSlFifoPtoU, CY_FX_SLFIFO_DMA_RX_SIZE);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PDmaChannelSetXfer Failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

#ifdef EXPLORE_GPIF_NOISE
    /* There were some LPM and performance related issues reported. Let's disable LPM for
      our specific test case. */
    CyU3PUsbLPMDisable();
#endif

    /* Update the status flag. */
    glIsApplnActive = CyTrue;
}

/* This function stops the slave FIFO loop application. This shall be called
 * whenever a RESET or DISCONNECT event is received from the USB host. The
 * endpoints are disabled and the DMA pipe is destroyed by this function. */
void
CyFxSlFifoApplnStop (
        void)
{
    CyU3PEpConfig_t epCfg;
    CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;

    /* Update the flag. */
    glIsApplnActive = CyFalse;

    /* Flush the endpoint memory */
    CyU3PUsbFlushEp(CY_FX_EP_PRODUCER);
    CyU3PUsbFlushEp(CY_FX_EP_CONSUMER);

    /* Destroy the channel */
    CyU3PDmaChannelDestroy (&glChHandleSlFifoUtoP);
    CyU3PDmaChannelDestroy (&glChHandleSlFifoPtoU);

    /* Disable endpoints. */
    CyU3PMemSet ((uint8_t *)&epCfg, 0, sizeof (epCfg));
    epCfg.enable = CyFalse;

    /* Producer endpoint configuration. */
    apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_PRODUCER, &epCfg);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "Stop Producer CyU3PSetEpConfig failed, Error code = %d\n", apiRetStatus);
       // CyFxAppErrorHandler (apiRetStatus);
    }

    /* Consumer endpoint configuration. */
    apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_CONSUMER, &epCfg);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "Stop Consumer CyU3PSetEpConfig failed, Error code = %d\n", apiRetStatus);
       // CyFxAppErrorHandler (apiRetStatus);
    }
}

/* Callback to handle the USB setup requests. */
CyBool_t
CyFxSlFifoApplnUSBSetupCB (
        uint32_t setupdat0,
        uint32_t setupdat1
    )
{
    /* Fast enumeration is used. Only requests addressed to the interface, class,
     * vendor and unknown control requests are received by this function.
     * This application does not support any class or vendor requests. */

    uint8_t  bRequest, bReqType;
    uint8_t  bType, bTarget;
    uint16_t wValue, wIndex;
    CyBool_t isHandled = CyFalse;

   // CyU3PDebugPrint (4, "Setup callback: %X %X\r\n", setupdat0, setupdat1 );

    /* Decode the fields from the setup request. */
    bReqType = (setupdat0 & CY_U3P_USB_REQUEST_TYPE_MASK);
    bType    = (bReqType & CY_U3P_USB_TYPE_MASK);
    bTarget  = (bReqType & CY_U3P_USB_TARGET_MASK);
    bRequest = ((setupdat0 & CY_U3P_USB_REQUEST_MASK) >> CY_U3P_USB_REQUEST_POS);
    wValue   = ((setupdat0 & CY_U3P_USB_VALUE_MASK)   >> CY_U3P_USB_VALUE_POS);
    wIndex   = ((setupdat1 & CY_U3P_USB_INDEX_MASK)   >> CY_U3P_USB_INDEX_POS);

    if (bType == CY_U3P_USB_STANDARD_RQT)
    {
        /* Handle SET_FEATURE(FUNCTION_SUSPEND) and CLEAR_FEATURE(FUNCTION_SUSPEND)
         * requests here. It should be allowed to pass if the device is in configured
         * state and failed otherwise. */
        if ((bTarget == CY_U3P_USB_TARGET_INTF) && ((bRequest == CY_U3P_USB_SC_SET_FEATURE)
                    || (bRequest == CY_U3P_USB_SC_CLEAR_FEATURE)) && (wValue == 0))
        {
            if (glIsApplnActive)
                CyU3PUsbAckSetup ();
            else
                CyU3PUsbStall (0, CyTrue, CyFalse);

            isHandled = CyTrue;
        }

        /* CLEAR_FEATURE request for endpoint is always passed to the setup callback
         * regardless of the enumeration model used. When a clear feature is received,
         * the previous transfer has to be flushed and cleaned up. This is done at the
         * protocol level. Since this is just a loopback operation, there is no higher
         * level protocol. So flush the EP memory and reset the DMA channel associated
         * with it. If there are more than one EP associated with the channel reset both
         * the EPs. The endpoint stall and toggle / sequence number is also expected to be
         * reset. Return CyFalse to make the library clear the stall and reset the endpoint
         * toggle. Or invoke the CyU3PUsbStall (ep, CyFalse, CyTrue) and return CyTrue.
         * Here we are clearing the stall. */
        if ((bTarget == CY_U3P_USB_TARGET_ENDPT) && (bRequest == CY_U3P_USB_SC_CLEAR_FEATURE)
                && (wValue == CY_U3P_USBX_FS_EP_HALT))
        {
            if (glIsApplnActive)
            {
                if (wIndex == CY_FX_EP_PRODUCER)
                {
                    CyU3PDmaChannelReset (&glChHandleSlFifoUtoP);
                    CyU3PUsbFlushEp(CY_FX_EP_PRODUCER);
                    CyU3PUsbResetEp (CY_FX_EP_PRODUCER);
                    CyU3PDmaChannelSetXfer (&glChHandleSlFifoUtoP, CY_FX_SLFIFO_DMA_TX_SIZE);
                }

                if (wIndex == CY_FX_EP_CONSUMER)
                {
                    CyU3PDmaChannelReset (&glChHandleSlFifoPtoU);
                    CyU3PUsbFlushEp(CY_FX_EP_CONSUMER);
                    CyU3PUsbResetEp (CY_FX_EP_CONSUMER);
                    CyU3PDmaChannelSetXfer (&glChHandleSlFifoPtoU, CY_FX_SLFIFO_DMA_RX_SIZE);
                }

                CyU3PUsbStall (wIndex, CyFalse, CyTrue);
                isHandled = CyTrue;
            }
        }
    }


#ifdef EXPLORE_GPIF_NOISE
    // read errors' counters:
    // Req Type = Vendor
    // Target = Device
    // Direction = IN (i.e. host read)
    // Req Code = 0x01
    // wValue = reset error counters flag  (0 - no reset; any other value - reset)
    // wIndex is ignored

    // Returns 21 bytes, all values are LE:
    // 64-bit SuperSpeed PHY errors count
    // 64-bit SuperSpeed LINK errors count,
    // 32-bit USB Disconnect/Reset count,
    // 8-bit USB speed mode
    if ( (bType == CY_U3P_USB_VENDOR_RQT)
    	&& (CY_U3P_USB_TARGET_DEVICE==bTarget)
    	&& (0x01==bRequest)
    	&& (0x80 & bReqType) //direction IN
    	)
		{
			uint16_t wLength   = ((setupdat1 & CY_U3P_USB_LENGTH_MASK)   >> CY_U3P_USB_LENGTH_POS);
			static uint8_t CacheAlignedBuffer[32] __attribute__ ((aligned (32))) ;
			static uint64_t curr_PhyErrorCount=0, init_PhyErrorCount=0;
			static uint64_t curr_LnkErrorCount=0, init_LnkErrorCount=0;

			uint8_t *Ep0Buffer=(uint8_t *)CacheAlignedBuffer;
			uint32_t e;

			// Historically I selected 64-bit counters to represent errors.
			// Let's continue to use this format for backward compatibility with PC software.
			e=glPhyErrorCount;
			if (e<(uint32_t)curr_PhyErrorCount)
			{
				// 32-bit overflow has appeared.
				curr_PhyErrorCount+=0x100000000ULL;
			}
			curr_PhyErrorCount=(curr_PhyErrorCount & 0xFFFFFFFF00000000ULL) | e;

			e=glLnkErrorCount;
			if (e<(uint32_t)curr_LnkErrorCount)
			{
				// 32-bit overflow has appeared.
				curr_LnkErrorCount+=0x100000000ULL;
			}
			curr_LnkErrorCount=(curr_LnkErrorCount & 0xFFFFFFFF00000000ULL) | e;

			*(uint64_t *)(Ep0Buffer+0) = curr_PhyErrorCount-init_PhyErrorCount;  //64-bit counter
			*(uint64_t *)(Ep0Buffer+8) = curr_LnkErrorCount-init_LnkErrorCount;  //64-bit counter
			*(uint32_t *)(Ep0Buffer+16)= glUSBDisconnectCount;      //32-bit counter
			Ep0Buffer[20] = CyU3PUsbGetSpeed(); // Error counters have meaning only in USB 3.0 SuperSpeed mode
			*(uint32_t *)(Ep0Buffer+24)=*(volatile uint32_t *)(0xe0033000+20); //FX3 hardware provided errors register
			*(uint32_t *)(Ep0Buffer+28)=glErrorUpdateCount; //32-bit counter, used for tracking whether the error counters get updated or not
			// if wValue!=0 then, as side effect, restart counting of PHY and LINK errors from zero
			if (wValue)
			{
				init_PhyErrorCount=curr_PhyErrorCount;
				init_LnkErrorCount=curr_LnkErrorCount;
			}

			if (wLength)
			{
				CyU3PUsbSendEP0Data(CY_U3P_MIN(wLength,/* 21*/32), (uint8_t *)Ep0Buffer);
			}
			else
			{
				CyU3PUsbAckSetup ();

			}
			// request is handled
			return CyTrue;
		}

#endif

    return isHandled;
}

/* This is the callback function to handle the USB events. */
void
CyFxSlFifoApplnUSBEventCB (
    CyU3PUsbEventType_t evtype,
    uint16_t            evdata
    )
{
    switch (evtype)
    {
    	case CY_U3P_USB_EVENT_LNK_RECOVERY:
    		// USB reset due to LINK errors
    		if (0xFFFF==evdata) glLinkErrLimitExceeded++;
    		else glLinkRecovery++;
    		break;
        case CY_U3P_USB_EVENT_SETCONF:
            /* Stop the application before re-starting. */
            if (glIsApplnActive)
            {
                CyFxSlFifoApplnStop ();
            }
            /* Start the loop back function. */
            CyFxSlFifoApplnStart ();
            break;

        case CY_U3P_USB_EVENT_RESET:
        case CY_U3P_USB_EVENT_DISCONNECT:
            /* Stop the loop back function. */
        	if (glIsApplnActive)
            {
                CyFxSlFifoApplnStop ();
            }
#ifdef EXPLORE_GPIF_NOISE
        	glUSBDisconnectCount++;
#endif
            break;

        default:
            break;
    }
}

/* Callback function to handle LPM requests from the USB 3.0 host. This function is invoked by the API
   whenever a state change from U0 -> U1 or U0 -> U2 happens. If we return CyTrue from this function, the
   FX3 device is retained in the low power state. If we return CyFalse, the FX3 device immediately tries
   to trigger an exit back to U0.

   This application does not have any state in which we should not allow U1/U2 transitions; and therefore
   the function always return CyTrue.
 */
CyBool_t
CyFxApplnLPMRqtCB (
        CyU3PUsbLinkPowerMode link_mode)
{
    return CyTrue;
}

/* This function initializes the GPIF interface and initializes
 * the USB interface. */
void
CyFxSlFifoApplnInit (void)
{
    CyU3PPibClock_t pibClock={0};
    CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;

#ifdef EXPLORE_GPIF_NOISE
    /* Initialize the p-port block. */
#if 10
    // the worst case for 403.2 MHz system clock
    // GPIF clock=403.2/5=80.64MHz
    pibClock.clkDiv = 5;
    pibClock.clkSrc = CY_U3P_SYS_CLK;
    pibClock.isHalfDiv = CyFalse;
#else
    //  the worst case for default (384 MHz) system clock seems to be
    //  GPIF clock=384/4.5=85.33MHz
    pibClock.clkDiv = 4;
    pibClock.clkSrc = CY_U3P_SYS_CLK;
    pibClock.isHalfDiv = CyTrue;
#endif
    /* Disable DLL for sync GPIF */
    pibClock.isDllEnable = CyFalse;
    apiRetStatus = CyU3PPibInit(CyTrue, &pibClock);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "P-port Initialization failed, Error Code = %d\n",apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Load the GPIF configuration for FX3 self-performing read/write state machine. */
    apiRetStatus = CyU3PGpifLoad (&Fx3gpif_Selfrdwr_CyFxGpifConfig);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PGpifLoad failed, Error Code = %d\n",apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Start the state machine. */
    apiRetStatus = CyU3PGpifSMStart (FX3GPIF_SELFRDWR_RESET, FX3GPIF_SELFRDWR_ALPHA_RESET);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PGpifSMStart failed, Error Code = %d\n",apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }
#else
    /* Initialize the p-port block. */
    pibClock.clkDiv = 2;
    pibClock.clkSrc = CY_U3P_SYS_CLK;
    pibClock.isHalfDiv = CyFalse;
    /* Disable DLL for sync GPIF */
    pibClock.isDllEnable = CyFalse;
    apiRetStatus = CyU3PPibInit(CyTrue, &pibClock);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "P-port Initialization failed, Error Code = %d\n",apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Load the GPIF configuration for Slave FIFO sync mode. */
    apiRetStatus = CyU3PGpifLoad (&Sync_Slave_Fifo_2Bit_CyFxGpifConfig);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PGpifLoad failed, Error Code = %d\n",apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Start the state machine. */
    apiRetStatus = CyU3PGpifSMStart (SYNC_SLAVE_FIFO_2BIT_RESET, SYNC_SLAVE_FIFO_2BIT_ALPHA_RESET);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PGpifSMStart failed, Error Code = %d\n",apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }
#endif

    /* Start the USB functionality. */


    apiRetStatus = CyU3PUsbStart();
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "CyU3PUsbStart failed to Start, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* The fast enumeration is the easiest way to setup a USB connection,
     * where all enumeration phase is handled by the library. Only the
     * class / vendor requests need to be handled by the application. */
    CyU3PUsbRegisterSetupCallback(CyFxSlFifoApplnUSBSetupCB, /*CyTrue*/ CyFalse);

    /* Setup the callback to handle the USB events. */
    CyU3PUsbRegisterEventCallback(CyFxSlFifoApplnUSBEventCB);


    /* Register a callback to handle LPM requests from the USB 3.0 host. */
#ifdef EXPLORE_GPIF_NOISE
    // We do not need LPM callback. We disable LPM anyway by calling
    // CyU3PUsbLPMDisable().
    // Less software callbacks -> less load on CPU/memory -> hopefully better performance
#else
    CyU3PUsbRegisterLPMRequestCallback(CyFxApplnLPMRqtCB);
#endif

    /* Set the USB Enumeration descriptors */

    /* Super speed device descriptor. */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_DEVICE_DESCR, NULL, (uint8_t *)CyFxUSB30DeviceDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB set device descriptor failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* High speed device descriptor. */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_HS_DEVICE_DESCR, NULL, (uint8_t *)CyFxUSB20DeviceDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB set device descriptor failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* BOS descriptor */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_BOS_DESCR, NULL, (uint8_t *)CyFxUSBBOSDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB set configuration descriptor failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Device qualifier descriptor */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_DEVQUAL_DESCR, NULL, (uint8_t *)CyFxUSBDeviceQualDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB set device qualifier descriptor failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Super speed configuration descriptor */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_CONFIG_DESCR, NULL, (uint8_t *)CyFxUSBSSConfigDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB set configuration descriptor failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* High speed configuration descriptor */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_HS_CONFIG_DESCR, NULL, (uint8_t *)CyFxUSBHSConfigDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB Set Other Speed Descriptor failed, Error Code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* Full speed configuration descriptor */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_FS_CONFIG_DESCR, NULL, (uint8_t *)CyFxUSBFSConfigDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB Set Configuration Descriptor failed, Error Code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* String descriptor 0 */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 0, (uint8_t *)CyFxUSBStringLangIDDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB set string descriptor failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* String descriptor 1 */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 1, (uint8_t *)CyFxUSBManufactureDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB set string descriptor failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }

    /* String descriptor 2 */
    apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 2, (uint8_t *)CyFxUSBProductDscr);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB set string descriptor failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }


#ifdef EXPLORE_GPIF_NOISE
    // To be sure that host has noticed device disconnect, sleep a bit before connecting again.
    CyU3PThreadSleep (100);
#endif
    /* Connect the USB Pins with super speed operation enabled. */
    apiRetStatus = CyU3PConnectState(CyTrue, CyTrue);
    if (apiRetStatus != CY_U3P_SUCCESS)
    {
        CyU3PDebugPrint (4, "USB Connect failed, Error code = %d\n", apiRetStatus);
        //CyFxAppErrorHandler(apiRetStatus);
    }
}

#ifdef EXPLORE_GPIF_NOISE
/*
 * NB! FX3 firmware (at least versions 1.2.2 and 1.2.3 and 1.3.1) clears hardware
 * Error Counter Register periodically in timer interrupt, specifically in
 * CyU3PUibLnkErrClrTimerCb function.
 * Also CyU3PUsbGetErrorCounts implementation itself clears the same hardware
 * register, causing this way concurrency between hardware and software updating
 * the same register. As a result, if both HW and SW update the register
 * concurrently, one of the operations has no effect on content of register.
 *
 * Function below is free of CyU3PUsbGetErrorCounts flaw. Unfortunately its
 * not possible to avoid hardware register clearing in timer interrupt, but
 * assuming that MyU3PUsbGetErrorCounts function is called frequently enough,
 * hopefully still a better accuracy can be achieved.
 *
 * Note 1. MyU3PUsbGetErrorCounts function is not multi-thread safe (as also
 *       CyU3PUsbGetErrorCounts is not).
 */
static CyU3PReturnStatus_t MyU3PUsbGetErrorCounts(uint16_t *phy_err_cnt, uint16_t *lnk_err_cnt)
{
	// This function does not update PHY/LINK errors register.
	// Instead, it preserves the previous register value and
	// calculates the changes based on previous and current counter
	// values.
	static uint32_t current_value=0; //NB! static variable preserves its value between function calls
	register uint32_t previous_value;
	register uint16_t delta;

	if (!phy_err_cnt || !lnk_err_cnt)
	{
		return CY_U3P_ERROR_BAD_ARGUMENT;
	}

	// Take the previous register value
	previous_value=current_value;
	// Read current PHY/LINK errors register value (address 0xe0033000+20)
    // bits 16...31 - PHY  errors
    // bits  0...15 - LINK errors
	{
		// CPU and USB state machine are in separate clock domains
		// and therefore CPU can see any arbitrary value when
		// USB state machine updates the register at the same moment.
		// Let's assume that if three read operations return
        // equal values then it must be correct value - it's
        // very very unlikely that it's not.
		register uint32_t r[3];
		r[0]=*(volatile uint32_t *)(0xe0033000+20);
		r[1]=*(volatile uint32_t *)(0xe0033000+20);
		r[2]=*(volatile uint32_t *)(0xe0033000+20);
		if (r[0]==r[1] && r[0]==r[2])
		{
			current_value=r[0];
		}
		else
		{
			// Three values were not equal.
			// Let's try to read again but at maximal speed without any possible pauses
			// between read operations. The assumption is that USB automata does not
			// increment error counters very frequently and therefore we should see at
			// least two identical values.
			register uint32_t m=disable_interrupts();
			r[0]=*(volatile uint32_t *)(0xe0033000+20);
			r[1]=*(volatile uint32_t *)(0xe0033000+20);
			r[2]=*(volatile uint32_t *)(0xe0033000+20);
			restore_interrupts(m);
			// Update current value depending on what the two values were identical.
			// If there are no identical values then let's assume that
			// "current value" is equal to "previous value" this time - hopefully next
			// time we will have better luck.
			if (r[0]==r[1]) current_value=r[0];
			else if (r[1]==r[2]) current_value=r[1];
		}
	}
	if (CY_U3P_SUPER_SPEED!=CyU3PUsbGetSpeed())
	{
		return CY_U3P_ERROR_INVALID_SEQUENCE;
	}

	// NB! According Cypress technical support information, hardware error counters saturate
    // at value 0xFFFF (they never turn around to 0). Therefore, FX3 API needs to clear
	// counters itself periodically (it does this typically in timer interrupt).

	// NB! FX3 API version 1.3.1 clears LNK error counter but does not clear
	//     PHY error counter in timer interrupt. Therefore we need to clear counters
	//     ourselves if they have saturated. But as this is dangerous operation (USB
	//     state machine may also read any arbitrary value while CPU updates register,
	//     plus, we will lose errors that have appeared between our read and this write)
	//     then do this only when this is needed indeed.
	if (((current_value & 0xFFFF0000) == 0xFFFF0000) || ((current_value & 0x0000FFFF) == 0x0000FFFF))
	{
		*(volatile uint32_t *)(0xe0033000+20)=0;
	}

	// Calculate PHY errors increment
	delta=(uint16_t)(current_value>>16);
	if ((uint16_t)(previous_value>>16) <= delta)
	{
		delta -= (uint16_t)(previous_value>>16);
	}
	*phy_err_cnt=delta;

	// Calculate LINK errors increment
	delta=(uint16_t)(current_value);
	if ((uint16_t)previous_value <= delta)
	{
		delta -= (uint16_t)previous_value;
	}
	*lnk_err_cnt=delta;

	return CY_U3P_SUCCESS;
}
#endif

/* Entry function for the slFifoAppThread. */
void
SlFifoAppThread_Entry (
        uint32_t input)
{
    /* Initialize the debug module */
    CyFxSlFifoApplnDebugInit();

    /* Initialize the slave FIFO application */
    CyFxSlFifoApplnInit();

#ifdef EXPLORE_GPIF_NOISE
    /*
     * There is an issue with error counters in FX3 firmware (at least in versions 1.2.2
     * and 1.2.3). Hardware updated error counter register gets cleared periodically by
     * firmware in timer interrupt. Therefore we need to read this register as frequently
     * as possible and we can't sleep.
     * Instead, read errors continuously but let's do this at lower priority.
     */
    {
    	// According CyU3PThreadCreate description, threads should take only
        // priority values from 7 (the highest) to 15 (the lowest).
    	unsigned int old_priority;
        CyU3PThreadPriorityChange(CyU3PThreadIdentify(),14,&old_priority);
    }
 #endif

    for (;;)
    {
#ifdef EXPLORE_GPIF_NOISE
    	// NB! Due to PHY/LINK error counters clearing issues in FX3 firmware versions
        // 1.2.3 and prior, we do not sleep. Instead we read error counters as
        // frequently as possible, minimizing this way the chance we miss errors.

    	{
    		// print alive text
    		static uint32_t cnt=0, n=0;
    		if (cnt++>1000000)
    		{
    			cnt=0;
    			n++;
     			CyU3PDebugPrint (4, "%d) Main thread alive\r\nglPhyErrorCount=%X, glLnkErrorCount=%X, HW reg=%X\r\n",
     						n, glPhyErrorCount, glLnkErrorCount, *(volatile uint32_t *)(0xe0033000+20) );
    		}
    	}

    	//CyU3PThreadSleep (1);
		{
        	uint16_t phy_err_cnt = 0;
			uint16_t lnk_err_cnt = 0;
			// NB! Use our own MyU3PUsbGetErrorCounts implementation that's free of
			// CyU3PUsbGetErrorCounts flaws.
			// CyU3PReturnStatus_t status = CyU3PUsbGetErrorCounts(&phy_err_cnt, &lnk_err_cnt);
			CyU3PReturnStatus_t status = MyU3PUsbGetErrorCounts(&phy_err_cnt, &lnk_err_cnt);
			if (CY_U3P_SUCCESS==status)
			{
				if  (phy_err_cnt | lnk_err_cnt)  // optimization: update counters only if there is a reason for that
				{
					glPhyErrorCount += phy_err_cnt;
					glLnkErrorCount += lnk_err_cnt;
				}

				//for debugging purposes, increment a counter each time the errors are checked/updated
				glErrorUpdateCount+=1;
			}
		}

		{
			static uint32_t prev_cnt=0;
			static uint32_t prev_recovery_cnt=0;

			if (prev_cnt!=glLinkErrLimitExceeded)
			{
				prev_cnt=glLinkErrLimitExceeded;
				CyU3PDebugPrint (4, "Link errors limit exceeded %d times!\r\n", prev_cnt);
			}
			if (prev_recovery_cnt!=glLinkRecovery)
			{
				prev_recovery_cnt=glLinkRecovery;
				CyU3PDebugPrint (4, "Link recovered %d times!\r\n", prev_recovery_cnt);
			}
		}
#else
        CyU3PThreadSleep (1000);
        if (glIsApplnActive)
        {
            /* Print the number of buffers received so far from the USB host. */
            CyU3PDebugPrint (6, "Data tracker: buffers received: %d, buffers sent: %d.\n",
                    glDMARxCount, glDMATxCount);
        }

#endif
    }
}

/* Application define function which creates the threads. */
void
CyFxApplicationDefine (
        void)
{
    void *ptr = NULL;
    uint32_t retThrdCreate = CY_U3P_SUCCESS;

    /* Allocate the memory for the thread */
    //ptr = CyU3PMemAlloc (CY_FX_SLFIFO_THREAD_STACK);
    // NB! align stack to 8-byte boundary
    static uint32_t stack[(CY_FX_SLFIFO_THREAD_STACK+7)/8*2+1];
    ptr=((uint32_t)stack & 0x07)?stack+1:stack;

    /* Create the thread for the application */
    retThrdCreate = CyU3PThreadCreate (&slFifoAppThread,           /* Slave FIFO app thread structure */
                          "21:Slave_FIFO_sync",                    /* Thread ID and thread name */
                          SlFifoAppThread_Entry,                   /* Slave FIFO app thread entry function */
                          0,                                       /* No input parameter to thread */
                          ptr,                                     /* Pointer to the allocated thread stack */
                          (CY_FX_SLFIFO_THREAD_STACK+7)/8*8,               /* App Thread stack size */
                          CY_FX_SLFIFO_THREAD_PRIORITY,            /* App Thread priority */
                          CY_FX_SLFIFO_THREAD_PRIORITY,            /* App Thread pre-emption threshold */
                          CYU3P_NO_TIME_SLICE,                     /* No time slice for the application thread */
                          CYU3P_AUTO_START                         /* Start the thread immediately */
                          );

    /* Check the return code */
    if (retThrdCreate != 0)
    {
        /* Thread Creation failed with the error code retThrdCreate */

        /* Add custom recovery or debug actions here */

        /* Application cannot continue */
        /* Loop indefinitely */
        while(1);
    }
}

/*
 * Main function
 */
int
main (void)
{
    CyU3PIoMatrixConfig_t io_cfg={0};
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;
    /* Initialize the device */
#ifdef EXPLORE_GPIF_NOISE

    // For max performance let's use 403.2 MHz clock (if external clock is 19.2MHz)
	CyU3PSysClockConfig_t sys_clock = {0};
	sys_clock.setSysClk400=CyTrue;
	sys_clock.cpuClkDiv=2;
	sys_clock.dmaClkDiv=2;
	sys_clock.mmioClkDiv=2;
	sys_clock.useStandbyClk=CyFalse;
	sys_clock.clkSrc=CY_U3P_SYS_CLK;

	status = CyU3PDeviceInit(&sys_clock);
#else
	// Default clock is 384 MHz (if external clock is 19.2MHz)
    status = CyU3PDeviceInit (NULL);
#endif
    if (status != CY_U3P_SUCCESS)
    {
       goto handle_fatal_error;
    }

    /* Initialize the caches. Enable both Instruction and Data Caches. */
    status = CyU3PDeviceCacheControl (CyTrue, CyTrue, CyTrue);
    if (status != CY_U3P_SUCCESS)
    {
        goto handle_fatal_error;
    }

    /* Configure the IO matrix for the device. On the FX3 DVK board, the COM port 
     * is connected to the IO(53:56). This means that either DQ32 mode should be
     * selected or lppMode should be set to UART_ONLY. Here we are choosing
     * UART_ONLY configuration for 16 bit slave FIFO configuration and setting
     * isDQ32Bit for 32-bit slave FIFO configuration. */
    io_cfg.useUart   = CyTrue;
    io_cfg.useI2C    = CyFalse;
    io_cfg.useI2S    = CyFalse;
    io_cfg.useSpi    = CyFalse;
#ifdef EXPLORE_GPIF_NOISE
    io_cfg.isDQ32Bit = CyTrue;
    io_cfg.lppMode   = CY_U3P_IO_MATRIX_LPP_DEFAULT;
#else

#if (CY_FX_SLFIFO_GPIF_16_32BIT_CONF_SELECT == 0)
    io_cfg.isDQ32Bit = CyFalse;
    io_cfg.lppMode   = CY_U3P_IO_MATRIX_LPP_UART_ONLY;
#else
    io_cfg.isDQ32Bit = CyTrue;
    io_cfg.lppMode   = CY_U3P_IO_MATRIX_LPP_DEFAULT;
#endif

#endif
    /* No GPIOs are enabled. */
    io_cfg.gpioSimpleEn[0]  = 0;
    io_cfg.gpioSimpleEn[1]  = 0;
    io_cfg.gpioComplexEn[0] = 0;
    io_cfg.gpioComplexEn[1] = 0;
    status = CyU3PDeviceConfigureIOMatrix (&io_cfg);
    if (status != CY_U3P_SUCCESS)
    {
        goto handle_fatal_error;
    }

    /* This is a non returnable call for initializing the RTOS kernel */
    CyU3PKernelEntry ();

    /* Dummy return to make the compiler happy */
    return 0;

handle_fatal_error:

    /* Cannot recover from this error. */
    while (1);
}

/* [ ] */

